from typing import List

from lightning.app.storage import Drive

from lightning_hpo.commands.tensorboard.stop import TensorboardConfig
from lightning_hpo.components.tensorboard import Tensorboard
from lightning_hpo.controllers.controller import Controller


class TensorboardController(Controller):

    model = TensorboardConfig

    def on_reconcile_start(self, configs: List[TensorboardConfig]):
        for config in configs:
            if config.sweep_id not in self.resources:
                self.resources[config.sweep_id] = Tensorboard(
                    component_name=config.sweep_id,
                    drive=Drive("lit://logs"),
                )



from argparse import ArgumentParser

from lightning.pytorch import Trainer
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.demos.boring_classes import BoringModel

if __name__ == "__main__":

    class BoringModelLogging(BoringModel):
        def validation_step(self, batch, batch_idx):
            res = super().validation_step(batch, batch_idx)
            self.log("val_loss", res["x"])
            return res

    parser = ArgumentParser()
    parser.add_argument("--max_epochs", type=float)
    args = parser.parse_args()

    model = BoringModelLogging()
    callback = ModelCheckpoint(save_last=True, monitor="val_loss")
    trainer = Trainer(max_epochs=int(args.max_epochs), limit_train_batches=5, callbacks=[callback])
    trainer.fit(model)



from argparse import ArgumentParser
from getpass import getuser
from typing import List, Optional
from uuid import uuid4

from lightning.app.utilities.commands import ClientCommand
from sqlalchemy import Column
from sqlmodel import Field, SQLModel

from lightning_hpo.utilities.enum import Status
from lightning_hpo.utilities.utils import pydantic_column_type


class NotebookConfig(SQLModel, table=True):

    __table_args__ = {"extend_existing": True}

    id: Optional[str] = Field(default=None, primary_key=True)
    name: str
    requirements: List[str] = Field(..., sa_column=Column(pydantic_column_type(List[str])))
    cloud_compute: str
    status: str = Status.NOT_STARTED
    desired_state: str = Status.RUNNING


class RunNotebookCommand(ClientCommand):
    def run(self) -> None:
        parser = ArgumentParser()

        parser.add_argument("name", help="The name of your notebook to run.")
        parser.add_argument("--requirements", nargs="+", default=[], help="Requirements file.")
        parser.add_argument("--cloud_compute", default="cpu", type=str, help="The machine to use in the cloud.")

        hparams, _ = parser.parse_known_args()
        id = str(uuid4()).split("-")[0]

        config = NotebookConfig(
            id=f"{getuser()}-{id}",
            name=hparams.name,
            requirements=hparams.requirements,
            cloud_compute=hparams.cloud_compute,
        )
        response = self.invoke_handler(config=config)
        print(response)



from typing import Optional, Type

import requests
from requests import Session
from requests.adapters import HTTPAdapter
from sqlmodel import SQLModel
from urllib3.util.retry import Retry

from lightning_hpo.components.servers.db.models import GeneralModel

_CONNECTION_RETRY_TOTAL = 5
_CONNECTION_RETRY_BACKOFF_FACTOR = 0.5


def _configure_session() -> Session:
    """Configures the session for GET and POST requests.

    It enables a generous retrial strategy that waits for the application server to connect.
    """
    retry_strategy = Retry(
        # wait time between retries increases exponentially according to: backoff_factor * (2 ** (retry - 1))
        total=_CONNECTION_RETRY_TOTAL,
        backoff_factor=_CONNECTION_RETRY_BACKOFF_FACTOR,
        status_forcelist=[429, 500, 502, 503, 504],
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    http = requests.Session()
    http.mount("https://", adapter)
    http.mount("http://", adapter)
    return http


class DatabaseConnector:
    def __init__(self, model: Type[SQLModel], db_url: str, model_id: Optional[str] = None):
        self.model = model
        self.db_url = db_url
        self.model_id = model_id
        self.session = _configure_session()

    def get(self, config: Optional[Type[SQLModel]] = None):
        cls = config if config else self.model
        resp = self.session.get(self.db_url, data=GeneralModel.from_cls(cls).json())
        assert resp.status_code == 200
        return [cls(**data) for data in resp.json()]

    def post(self, config: SQLModel, model_id: Optional[str] = None):
        resp = self.session.post(
            self.db_url,
            data=GeneralModel.from_obj(config, id=model_id or self.model_id).json(),
        )
        assert resp.status_code == 200

    def put(self, config: SQLModel, model_id: Optional[str] = None):
        resp = self.session.put(
            self.db_url,
            data=GeneralModel.from_obj(config, id=self.model_id or self.model_id).json(),
        )
        assert resp.status_code == 200

    def delete(self, config: SQLModel, model_id: Optional[str] = None):
        resp = self.session.delete(
            self.db_url,
            data=GeneralModel.from_obj(config, id=self.model_id or self.model_id).json(),
        )
        assert resp.status_code == 200



from lightning import LightningFlow
from lightning.app.storage import Drive

from lightning_hpo.commands.artefacts.download import (
    _collect_artefact_urls,
    DownloadArtefactsCommand,
    DownloadArtefactsConfig,
)
from lightning_hpo.commands.artefacts.show import _collect_artefact_paths, ShowArtefactsCommand, ShowArtefactsConfig
from lightning_hpo.components.servers.db import Database, DatabaseViz
from lightning_hpo.components.servers.file_server import FileServer
from lightning_hpo.controllers.notebook import NotebookController
from lightning_hpo.controllers.sweep import SweepController
from lightning_hpo.controllers.tensorboard import TensorboardController
from lightning_hpo.utilities.enum import Status


class MainFlow(LightningFlow):
    def __init__(self, debug: bool = False):
        super().__init__()
        self.debug = debug

        # 1: Create Drive
        self.drive = Drive("lit://code")

        # 2: Controllers
        self.sweep_controller = SweepController(self.drive)
        self.notebook_controller = NotebookController()
        self.tensorboard_controller = TensorboardController()

        # 3: Create the File Server to upload code or data.
        self.file_server = FileServer(self.drive)

        # 4: Create the database.
        self.db = Database(
            models=[
                self.sweep_controller.model,
                self.notebook_controller.model,
                self.tensorboard_controller.model,
            ]
        )

        if self.debug:
            self.db_viz = DatabaseViz()

        self.ready = False

    def run(self):
        # 1: Start the servers.
        self.file_server.run()
        self.db.run()
        if self.debug:
            self.db_viz.run()

        # 2: Wait for the servers to be alive
        if not (self.file_server.alive() and self.db.alive()):
            return

        if not self.ready:
            print(f"The Training App is ready ! Database URL: {self.db.db_url}")
            self.ready = True

        # 3: Run the controllers
        self.sweep_controller.run(self.db.db_url)
        self.notebook_controller.run(self.db.db_url)
        self.tensorboard_controller.run(self.db.db_url)

    def configure_layout(self):
        tabs = [{"name": "Dashboard", "content": self.sweep_controller}]
        if self.debug:
            tabs += [{"name": "Database Viz", "content": self.db_viz}]
        for sweep in self.sweep_controller.resources.values():
            if sweep.show:
                tabs += sweep.configure_layout()

        for sweep_id, tensorboard in self.tensorboard_controller.resources.items():
            tabs += [{"name": f"tensorboard_{sweep_id}", "content": tensorboard}]

        for notebook_name, notebook in self.notebook_controller.resources.items():
            if notebook._config.desired_state == Status.RUNNING:
                tabs += [{"name": notebook_name, "content": notebook}]
        return tabs

    def show_artefacts(self, config: ShowArtefactsConfig):
        return _collect_artefact_paths(config)

    def download_artefacts(self, config: DownloadArtefactsConfig):
        return _collect_artefact_urls(config)

    def configure_commands(self):
        controller_commands = self.sweep_controller.configure_commands()
        controller_commands += self.notebook_controller.configure_commands()
        controller_commands += [
            {"show artefacts": ShowArtefactsCommand(self.show_artefacts)},
            {"download artefacts": DownloadArtefactsCommand(self.download_artefacts)},
        ]
        return controller_commands



import uuid
from typing import Any, Dict, List, Optional, Type, Union

from lightning import BuildConfig, CloudCompute, LightningFlow
from lightning.app.components.python.tracer import Code
from lightning.app.storage.path import Path

from lightning_hpo.algorithm.base import Algorithm
from lightning_hpo.algorithm.optuna import OptunaAlgorithm
from lightning_hpo.commands.sweep.run import Params, SweepConfig, TrialConfig
from lightning_hpo.distributions import Distribution
from lightning_hpo.distributions.distributions import parse_distributions, unparse_distributions
from lightning_hpo.framework.agnostic import Objective
from lightning_hpo.loggers import LoggerType
from lightning_hpo.utilities.enum import Status
from lightning_hpo.utilities.utils import (
    _check_status,
    _resolve_objective_cls,
    get_best_model_path,
    get_best_model_score,
    HPOCloudCompute,
)


class Sweep(LightningFlow):
    def __init__(
        self,
        n_trials: int,
        objective_cls: Optional[Type[Objective]] = None,
        simultaneous_trials: int = 1,
        script_args: Optional[Union[list, str]] = None,
        env: Optional[Dict] = None,
        cloud_compute: Optional[HPOCloudCompute] = None,
        script_path: Optional[str] = None,
        algorithm: Optional[Algorithm] = None,
        logger: str = "streamlit",
        sweep_id: Optional[str] = None,
        distributions: Optional[Dict[str, Distribution]] = None,
        framework: str = "base",
        code: Optional[Code] = None,
        direction: Optional[str] = None,
        trials_done: Optional[int] = 0,
        requirements: Optional[List[str]] = None,
        trials: Optional[Dict[int, TrialConfig]] = None,
        status: Optional[str] = Status.NOT_STARTED,
        **objective_kwargs: Any,
    ):
        """The Sweep class enables to easily run a Python Script with Lightning
        :class:`~lightning.utilities.tracer.Tracer` with state-of-the-art distributed.
        Arguments:
            n_trials: Number of HPO trials to run.
            objective_cls: Your custom base objective work.
            simultaneous_trials: Number of parallel trials to run.
            script_args: Optional script arguments.
            env: Environment variables to be passed to the script.
            cloud_compute: The cloud compute on which the Work should run on.
            blocking: Whether the Work should be blocking or asynchronous.
            script_path: Path of the python script to run.
            logger: Which logger to use
            objective_kwargs: Your custom keywords arguments passed to your custom objective work class.
        """
        super().__init__()
        # Sweep Database Spec
        self._sweep_config = SweepConfig(
            sweep_id=sweep_id or str(uuid.uuid4()).split("-")[0],
            script_path=script_path,
            n_trials=n_trials,
            simultaneous_trials=simultaneous_trials,
            trials_done=trials_done,
            requirements=requirements or [],
            script_args=script_args,
            distributions=unparse_distributions(distributions),
            framework=framework,
            cloud_compute=getattr(cloud_compute, "name", "default"),
            num_nodes=getattr(cloud_compute, "count", 1) if cloud_compute else 1,
            logger=logger,
            direction=direction,
            trials=trials or {},
            status=status,
        )

        self._objective_cls = _resolve_objective_cls(objective_cls, framework)

        self._algorithm = algorithm or OptunaAlgorithm(direction=direction)
        self._logger = LoggerType(logger).get_logger()
        self._logger.connect(self)

        self._kwargs = {
            "script_path": script_path,
            "env": env,
            "script_args": script_args,
            "cloud_compute": CloudCompute(name=cloud_compute.name if cloud_compute else "cpu"),
            "num_nodes": getattr(cloud_compute, "count", 1) if cloud_compute else 1,
            "logger": logger,
            "code": code,
            "sweep_id": self._sweep_config.sweep_id,
            "raise_exception": False,
            **objective_kwargs,
        }
        self._algorithm.register_distributions(
            {k: d.to_dict() if isinstance(d, Distribution) else d for k, d in (distributions or {}).items()}
        )
        self._algorithm.register_trials([t for t in trials.values() if t.status == Status.SUCCEEDED] if trials else [])
        self.restart_count = 0
        self.has_updated = False
        self.sweep = False
        self.show = False

    def run(self):
        if self._sweep_config.status in (Status.FAILED, Status.SUCCEEDED, Status.STOPPED):
            return

        if self._sweep_config.trials_done == self._sweep_config.n_trials:
            self._sweep_config.status = Status.SUCCEEDED
            self.has_updated = True
            return

        for trial_id in range(self._sweep_config.num_trials):

            objective = self._get_objective(trial_id)

            if objective:

                if _check_status(objective, Status.NOT_STARTED):
                    self._algorithm.trial_start(trial_id)
                    self._logger.on_after_trial_start(self._sweep_config.sweep_id)

                if not self._sweep_config.trials[trial_id].params.params:
                    self._sweep_config.status = Status.RUNNING
                    self._sweep_config.trials[trial_id].params = Params(params=self._algorithm.get_params(trial_id))
                    self.has_updated = True

                objective.run(
                    params=self._algorithm.get_params(trial_id),
                    restart_count=self.restart_count,
                )

                if _check_status(objective, Status.FAILED):
                    self._sweep_config.status = Status.FAILED
                    self._sweep_config.trials[trial_id].status = Status.FAILED
                    self.has_updated = True

                if objective.reports and not self._sweep_config.trials[trial_id].pruned:
                    if self._algorithm.should_prune(trial_id, objective.reports):
                        self._sweep_config.trials[trial_id].status = Status.PRUNED
                        self.has_updated = True
                        objective.stop()
                        continue

                if objective.best_model_score and not objective.has_stopped and not objective.pruned:
                    self._algorithm.trial_end(trial_id, objective.best_model_score)
                    self._logger.on_after_trial_end(
                        sweep_id=self._sweep_config.sweep_id,
                        trial_id=objective.trial_id,
                        monitor=objective.monitor,
                        score=objective.best_model_score,
                        params=self._algorithm.get_params(trial_id),
                    )
                    self._sweep_config.trials[trial_id].best_model_score = objective.best_model_score
                    self._sweep_config.trials[trial_id].best_model_path = objective.best_model_path
                    self._sweep_config.trials[trial_id].monitor = objective.monitor
                    self._sweep_config.trials[trial_id].status = Status.SUCCEEDED
                    self._sweep_config.trials_done += 1
                    self.has_updated = True
                    objective.stop()

    @property
    def best_model_score(self) -> Optional[float]:
        return get_best_model_score(self)

    @property
    def best_model_path(self) -> Optional[Path]:
        return get_best_model_path(self)

    def configure_layout(self):
        return self._logger.configure_layout()

    def _get_objective(self, trial_id: int):
        trial_config = self._sweep_config.trials.get(trial_id, None)
        if trial_config is None:
            trial_config = TrialConfig(
                best_model_score=None,
                monitor=None,
                best_model_path=None,
                status=Status.PENDING,
                params=Params(params={}),
            )
            self._sweep_config.trials[trial_id] = trial_config

        if trial_config.status == Status.SUCCEEDED:
            return

        objective = getattr(self, f"w_{trial_id}", None)
        if objective is None:
            objective = self._objective_cls(trial_id=trial_id, **self._kwargs)
            setattr(self, f"w_{trial_id}", objective)
        return objective

    @property
    def updates(self) -> List[SweepConfig]:
        res = []
        if self.has_updated:
            res = [self._sweep_config]
        self.has_updated = False
        return res

    @classmethod
    def from_config(cls, config: SweepConfig, code: Optional[Code] = None):
        return cls(
            script_path=config.script_path,
            n_trials=config.n_trials,
            simultaneous_trials=config.simultaneous_trials,
            framework=config.framework,
            script_args=config.script_args,
            trials_done=config.trials_done,
            distributions=parse_distributions(config.distributions),
            cloud_compute=HPOCloudCompute(config.cloud_compute, config.num_nodes),
            sweep_id=config.sweep_id,
            code=code,
            cloud_build_config=BuildConfig(requirements=config.requirements),
            logger=config.logger,
            algorithm=OptunaAlgorithm(direction=config.direction),
            trials=config.trials,
            direction=config.direction,
            status=config.status,
        )



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, TypedDict

from lightning_hpo.commands.sweep.run import Distributions, Params


class DistributionDict(TypedDict):
    distribution: str
    params: Dict[str, Any]


class Distribution(ABC):
    @abstractmethod
    def to_dict(self) -> DistributionDict:
        ...


class LogUniform(Distribution):
    def __init__(self, low: float, high: float) -> None:
        self.low = low
        self.high = high

    def to_dict(self) -> DistributionDict:
        return {"distribution": "log_uniform", "params": {"low": self.low, "high": self.high}}


class Uniform(Distribution):
    def __init__(self, low: float, high: float) -> None:
        self.low = low
        self.high = high

    def to_dict(self):
        return {"distribution": "uniform", "params": {"low": self.low, "high": self.high}}


class Categorical(Distribution):
    def __init__(self, choices: List[Any]) -> None:
        self.choices = choices

    def to_dict(self) -> DistributionDict:
        return {"distribution": "categorical", "params": {"choices": self.choices}}


class IntUniform(Distribution):
    def __init__(self, low: int, high: int, step: Optional[int] = 1) -> None:
        self.low = low
        self.high = high
        self.step = step

    def to_dict(self) -> DistributionDict:
        return {"distribution": "int_uniform", "params": {"low": self.low, "high": self.high, "step": self.step}}


_DISTRIBUTION = {"uniform": Uniform, "int_uniform": IntUniform, "log_uniform": LogUniform, "categorical": Categorical}


def parse_distributions(distributions: Dict[str, Distributions]) -> Dict[str, Distribution]:
    return {k: _DISTRIBUTION[v.distribution](**v.params.params) for k, v in distributions.items()}


def unparse_distributions(distributions: Optional[Dict[str, Distribution]]) -> Dict[str, Distributions]:
    if not distributions:
        return {}

    dist = {
        k: Distributions(distribution=x.to_dict()["distribution"], params=Params(params=x.to_dict()["params"]))
        for k, x in distributions.items()
    }
    return dist



import os
from time import time
from typing import Any, Dict, Optional

from lightning import LightningFlow
from lightning.app.storage import Drive
from lightning.pytorch.loggers import TensorBoardLogger
from lightning.pytorch.utilities.rank_zero import rank_zero_only

from lightning_hpo.loggers.logger import Logger


class DriveTensorBoardLogger(TensorBoardLogger):
    def __init__(self, *args, drive: Drive, refresh_time: int = 5, **kwargs):
        super().__init__(*args, **kwargs)
        self.timestamp = None
        self.drive = drive
        self.refresh_time = refresh_time

    @rank_zero_only
    def log_metrics(self, metrics, step) -> None:
        super().log_metrics(metrics, step)
        if self.timestamp is None:
            self.drive.put(self.log_dir)
            self.timestamp = time()
        elif (time() - self.timestamp) > self.refresh_time:
            self.drive.put(self.log_dir)
            self.timestamp = time()

    @rank_zero_only
    def finalize(self, status: str) -> None:
        super().finalize(status)
        self.drive.put(self.log_dir)


class TensorboardLogger(Logger):
    def on_after_trial_start(self, sweep_id: str, title: Optional[str] = None, desc: Optional[str] = None):
        pass

    def on_after_trial_end(self, sweep_id: str, trial_id: int, monitor: str, score: float, params: Dict[str, Any]):
        pass

    def connect(self, flow: LightningFlow):
        pass

    def configure_layout(self):
        return []

    def configure_tracer(self, tracer, sweep_id: str, trial_id: int, params: Dict[str, Any]):
        from pytorch_lightning import Trainer

        # Create a space logs under the sweep_id folder
        drive = Drive("lit://logs", component_name=f"{sweep_id}/{trial_id}")
        use_localhost = "LIGHTNING_APP_STATE_URL" not in os.environ

        if use_localhost:
            logger = TensorBoardLogger(save_dir=str(drive.root))
        else:
            logger = DriveTensorBoardLogger(save_dir=".", name="", drive=drive, refresh_time=5)

        print("Injecting Tensorboard")

        logger.log_hyperparams(params)

        def trainer_pre_fn(trainer, *args, **kwargs):
            kwargs["logger"] = logger
            return {}, args, kwargs

        tracer.add_traced(Trainer, "__init__", pre_fn=trainer_pre_fn)

    def get_url(self, trial_id: int) -> None:
        pass



import os
import re
from argparse import ArgumentParser
from getpass import getuser
from pathlib import Path
from typing import Dict, List, Optional, Union
from uuid import uuid4

import requests
from lightning.app.source_code import LocalSourceCodeDir
from lightning.app.source_code.uploader import FileUploader
from lightning.app.utilities.commands import ClientCommand
from sqlalchemy import Column
from sqlmodel import Field, SQLModel

from lightning_hpo.utilities.enum import Status
from lightning_hpo.utilities.utils import pydantic_column_type


class Params(SQLModel, table=False):
    params: Dict[str, Union[float, int, List[float], List[str]]] = Field(
        sa_column=Column(pydantic_column_type(Dict[str, Union[float, int, List[float], List[str]]]))
    )


class Distributions(SQLModel, table=False):
    distribution: str
    params: Params = Field(sa_column=Column(pydantic_column_type(Params)))


class TrialConfig(SQLModel, table=False):
    best_model_score: Optional[float]
    monitor: Optional[str]
    best_model_path: Optional[str]
    status: str = Status.NOT_STARTED
    params: Params = Field(sa_column=Column(pydantic_column_type(Params)))
    exception: Optional[str]

    @property
    def pruned(self) -> bool:
        return self.status == Status.PRUNED


class SweepConfig(SQLModel, table=True):
    __table_args__ = {"extend_existing": True}

    id: Optional[int] = Field(default=None, primary_key=True)
    sweep_id: str
    script_path: str
    n_trials: int
    simultaneous_trials: int
    trials_done: int = 0
    requirements: List[str] = Field(..., sa_column=Column(pydantic_column_type(List[str])))
    script_args: List[str] = Field(..., sa_column=Column(pydantic_column_type(List[str])))
    distributions: Dict[str, Distributions] = Field(
        ..., sa_column=Column(pydantic_column_type(Dict[str, Distributions]))
    )
    url: Optional[str] = None
    trials: Dict[int, TrialConfig] = Field(..., sa_column=Column(pydantic_column_type(Dict[int, TrialConfig])))
    framework: str
    cloud_compute: str
    num_nodes: int = 1
    logger: str
    direction: str
    status: str = Status.NOT_STARTED

    @property
    def num_trials(self) -> int:
        return self.trials_done + self.simultaneous_trials

    @property
    def username(self) -> str:
        return self.sweep_id.split("-")[0]

    @property
    def hash(self) -> str:
        return self.sweep_id.split("-")[1]


class DistributionParser:
    @staticmethod
    def is_distribution(argument: str) -> bool:
        ...

    @staticmethod
    def parse(argument: str) -> Dict:
        ...


class UniformDistributionParser(DistributionParser):
    @staticmethod
    def is_distribution(argument: str) -> bool:
        return "uniform" in argument

    @staticmethod
    def parse(argument: str):
        name, value = argument.split("=")
        regex = "[0-9]*\.[0-9]*"  # noqa W605
        low, high = re.findall(regex, value)
        return {name: {"distribution": "uniform", "params": {"low": float(low), "high": float(high)}}}


class LogUniformDistributionParser(DistributionParser):
    @staticmethod
    def is_distribution(argument: str) -> bool:
        return "log_uniform" in argument

    @staticmethod
    def parse(argument: str):
        name, value = argument.split("=")
        regex = "[0-9]*\.[0-9]*"  # noqa W605
        low, high = re.findall(regex, value)
        return {name: {"distribution": "log_uniform", "params": {"low": float(low), "high": float(high)}}}


class CategoricalDistributionParser(DistributionParser):
    @staticmethod
    def is_distribution(argument: str) -> bool:
        return "categorical" in argument

    @staticmethod
    def parse(argument: str):
        name, value = argument.split("=")
        choices = value.split("[")[1].split("]")[0].split(", ")
        return {name: {"distribution": "categorical", "params": {"choices": choices}}}


class CustomFileUploader(FileUploader):
    def _upload_data(self, s: requests.Session, url: str, data: bytes):
        resp = s.put(url, files={"file": data})
        return resp.status_code


class CustomLocalSourceCodeDir(LocalSourceCodeDir):
    def upload(self, url: str) -> None:
        """Uploads package to URL, usually pre-signed URL.

        Notes
        -----
        Since we do not use multipart uploads here, we cannot upload any
        packaged repository files which have a size > 2GB.

        This limitation should be removed during the datastore upload redesign
        """
        if self.package_path.stat().st_size > 2e9:
            raise OSError(
                "cannot upload directory code whose total fize size is greater than 2GB (2e9 bytes)"
            ) from None

        uploader = CustomFileUploader(
            presigned_url=url,
            source_file=str(self.package_path),
            name=self.package_path.name,
            total_size=self.package_path.stat().st_size,
        )
        uploader.upload()


class RunSweepCommand(ClientCommand):

    DESCRIPTION = "Command to run a Sweep or Trial"

    SUPPORTED_DISTRIBUTIONS = ("uniform", "log_uniform", "categorical")

    def run(self) -> None:
        parser = ArgumentParser()
        parser.add_argument("script_path", type=str, help="The path to the script to run.")
        parser.add_argument("--n_trials", type=int, help="Number of trials to run.")
        parser.add_argument("--simultaneous_trials", default=1, type=int, help="Number of trials to run.")
        parser.add_argument("--requirements", nargs="+", default=[], help="Requirements file.")
        parser.add_argument("--framework", default="pytorch_lightning", type=str, help="The framework you are using.")
        parser.add_argument("--cloud_compute", default="cpu", type=str, help="The machine to use in the cloud.")
        parser.add_argument("--sweep_id", default=None, type=str, help="The sweep you want to run upon.")
        parser.add_argument("--num_nodes", default=1, type=int, help="The number of nodes to train upon.")
        parser.add_argument("--logger", default="streamlit", type=str, help="The logger to use with your sweep.")
        parser.add_argument(
            "--direction",
            default="minimize",
            choices=["minimize", "maximize"],
            type=str,
            help="In which direction to optimize.",
        )
        hparams, args = parser.parse_known_args()

        if any("=" not in arg for arg in args):
            raise Exception("Please, provide the arguments as follows --x=y")

        script_args = []
        distributions = {}
        for arg in args:
            is_distribution = False
            for p in [UniformDistributionParser, LogUniformDistributionParser, CategoricalDistributionParser]:
                if p.is_distribution(arg.replace("--", "")):
                    distributions.update(p.parse(arg.replace("--", "")))
                    is_distribution = True
                    break
            if not is_distribution:
                script_args.append(arg)

        id = str(uuid4()).split("-")[0]
        sweep_id = hparams.sweep_id or f"{getuser()}-{id}"

        if not os.path.exists(hparams.script_path):
            raise Exception("The provided script doesn't exists.")

        repo = CustomLocalSourceCodeDir(path=Path(hparams.script_path).parent.resolve())
        # TODO: Resolve this bug.
        url = self.state.file_server._state["vars"]["_url"]
        repo.package()
        repo.upload(url=f"{url}/uploadfile/{sweep_id}")

        distributions = {
            k: Distributions(distribution=x["distribution"], params=Params(params=x["params"]))
            for k, x in distributions.items()
        }

        config = SweepConfig(
            sweep_id=sweep_id,
            script_path=hparams.script_path,
            n_trials=int(hparams.n_trials),
            simultaneous_trials=hparams.simultaneous_trials,
            requirements=hparams.requirements,
            script_args=script_args,
            distributions=distributions,
            framework=hparams.framework,
            cloud_compute=hparams.cloud_compute,
            num_nodes=hparams.num_nodes,
            logger=hparams.logger,
            direction=hparams.direction,
            trials={},
        )
        response = self.invoke_handler(config=config)
        print(response)



import argparse
import os
import re
from typing import Dict, List, Optional

import rich
from lightning.app.storage.path import filesystem, shared_storage_path
from lightning.app.utilities.commands import ClientCommand
from pydantic import BaseModel
from rich.text import Text
from rich.tree import Tree


def _walk_folder(tree: Tree, sorted_directories: Dict[str, List[str]], root_path: str, depth: int):
    parents = {}
    for directory in sorted_directories:
        splits = directory.split("/")
        if depth == len(splits):
            for path in sorted_directories[directory]:
                text_filename = Text(path.replace(directory + "/", ""), "green")
                text_filename.highlight_regex(r"\..*$", "bold red")
                text_filename.stylize(f"link {path}")
                icon = "🐍 " if path.endswith(".py") else "📄 "
                tree.add(Text(icon) + text_filename)
        else:
            root_folder = "/".join(splits[1 : depth + 1])  # noqa E203
            if root_folder not in parents:
                parents[root_folder] = []
            parents[root_folder].append(directory)

    if not parents:
        return

    for root_folder, directories in parents.items():
        folder = root_folder.split("/")[-1]
        style = "dim" if root_folder.startswith("__") else ""
        branch = tree.add(
            f"[bold magenta]:open_file_folder: {folder}",
            style=style,
            guide_style=style,
        )
        _walk_folder(
            branch,
            {directory: sorted_directories[directory] for directory in directories},
            os.path.join(root_path, folder),
            depth + 1,
        )


def walk_folder(paths: List[str], tree: Tree) -> None:
    """Recursively build a Tree with directory contents."""
    paths = sorted(paths)
    directories = {}
    for p in paths:
        directory = os.path.dirname(p)
        if directory not in directories:
            directories[directory] = []
        directories[directory].append(p)

    _walk_folder(tree, directories, "artifacts", 1)
    rich.print(tree)


class ShowArtefactsConfig(BaseModel):
    include: Optional[str] = None
    exclude: Optional[str] = None


class ShowArtefactsCommand(ClientCommand):
    def run(self) -> None:
        # 1. Parse the user arguments.
        parser = argparse.ArgumentParser()
        parser.add_argument("--include", type=str, default=None, help="Provide a regex to include some specific files.")
        parser.add_argument("--exclude", type=str, default=None, help="Provide a regex to exclude some specific files.")
        hparams = parser.parse_args()

        config = ShowArtefactsConfig(include=hparams.include, exclude=hparams.exclude)
        response: List[str] = self.invoke_handler(config=config)

        tree = Tree(
            "[bold magenta]:open_file_folder: root",
            guide_style="bold bright_blue",
        )

        walk_folder(response, tree)


def _filter_paths(paths: List[str], include: Optional[str], exclude: Optional[str]) -> List[str]:
    out = []

    if include:
        include_pattern = re.compile(include)

    if exclude:
        exclude_pattern = re.compile(exclude)

    for path in paths:
        if include or exclude:
            if exclude and len(exclude_pattern.findall(path)):
                pass

            elif (include and len(include_pattern.findall(path))) or include is None:
                out.append(path)
        else:
            out.append(path)
    return out


def _collect_artefact_paths(config: ShowArtefactsConfig, replace: bool = True) -> List[str]:
    """This function is responsible to collecting the files from the shared filesystem."""
    fs = filesystem()
    paths = []

    shared_storage = shared_storage_path()
    for root_dir, _, files in fs.walk(shared_storage):
        if replace:
            root_dir = str(root_dir).replace(str(shared_storage), "").replace("/artifacts/root.", "root/")
        for f in files:
            paths.append(os.path.join(str(root_dir), f))

    return _filter_paths(paths, config.include, config.exclude)



from abc import abstractmethod
from typing import List, Optional, Type

from lightning import LightningFlow
from lightning.app.storage import Drive
from lightning.app.structures import Dict
from sqlmodel import SQLModel

from lightning_hpo.components.servers.db import DatabaseConnector


class Controller(LightningFlow):

    model: Type[SQLModel]
    model_id: Optional[str] = None

    def __init__(self, drive: Optional[Drive] = None):
        super().__init__()
        self.db_url = None
        self.resources = Dict()
        self.drive = drive
        self._database = None

    def run(self, db_url: str, configs: Optional[List[Type[SQLModel]]] = None):
        self.db_url = db_url

        # TODO: Resolve scheduling. It seems only the last one is activated somehow in the cloud.
        # TODO: Improve the schedule API.
        # 1: Read from the database and generate the works accordingly.
        # if self.schedule("* * * * * 0,5,10,15,20,25,30,35,40,45,50,55"):
        db_configs = self.db.get()
        if configs:
            db_configs += db_configs
        if db_configs:
            self.on_reconcile_start(db_configs)

        # 2: Iterate over the resources and collect updates
        updates = []
        for resource in self.resources.values():
            resource.run()
            updates.extend(resource.updates)

        if not updates:
            return

        # 3: Reconcile resources on end
        for update in updates:
            self.db.put(update)

        self.on_reconcile_end(updates)

    @property
    def db(self) -> DatabaseConnector:
        if self._database is None:
            assert self.db_url is not None
            self._database = DatabaseConnector(self.model, self.db_url + "/general/", self.model_id or "id")
        return self._database

    @abstractmethod
    def on_reconcile_start(self, configs):
        """Override with your start reconciliation mechanism."""

    @abstractmethod
    def on_reconcile_end(self, updates):
        """Override with your end reconciliation mechanism."""



import inspect
import sys
from typing import Optional

from pydantic import BaseModel
from sqlmodel import Field, SQLModel


class GeneralModel(BaseModel):
    cls_name: str
    cls_module: str
    data: str
    id: Optional[str]

    def convert_to_model(self):
        return self.data_cls.parse_raw(self.data)

    @property
    def data_cls(self) -> BaseModel:
        return getattr(sys.modules[self.cls_module], self.cls_name)

    @classmethod
    def from_obj(cls, obj, id: Optional[str] = None):
        return cls(
            **{
                "cls_path": inspect.getfile(obj.__class__),
                "cls_name": obj.__class__.__name__,
                "cls_module": obj.__class__.__module__,
                "data": obj.json(),
                "id": id,
            }
        )

    @classmethod
    def from_cls(cls, obj_cls, id: Optional[str] = None):
        return cls(
            **{
                "cls_path": inspect.getfile(obj_cls),
                "cls_name": obj_cls.__name__,
                "cls_module": obj_cls.__module__,
                "data": "",
                "id": id,
            }
        )


class Trial(SQLModel, table=False):
    id: Optional[int] = Field(default=None, primary_key=True)
    sweep_id: str = Field(index=True)
    trial_id: int
    best_model_score: Optional[float]
    monitor: Optional[str]
    best_model_path: Optional[str]
    name: str
    has_succeeded: bool
    url: Optional[str]
    params: Optional[str]



import json
import logging
import os
import tarfile
import traceback
import uuid
import zipfile
from pathlib import Path

import lightning as L
from lightning.app.storage import Drive

_logger = logging.getLogger(__name__)


class FileServer(L.LightningWork):
    def __init__(self, drive: Drive, base_dir: str = "", chunk_size=10240, **kwargs):
        """This component uploads, downloads files to your application.

        Arguments:
            drive: The drive can share data inside your application.
            base_dir: The local directory where the data will be stored.
            chunk_size: The quantity of bytes to download/upload at once.
        """
        super().__init__(
            cloud_build_config=L.BuildConfig(requirements=["Flask", "Flask-Cors"]),
            parallel=True,
            **kwargs,
        )
        # 1: Attach the arguments to the state.
        self.drive = drive
        self.base_dir = base_dir
        self.chunk_size = chunk_size

        # 2: Create a folder to store the data.
        if self.base_dir:
            os.makedirs(self.base_dir, exist_ok=True)

        # 3: Keep a reference to the uploaded filenames.
        self.uploaded_files = dict()

    def get_filepath(self, path: str) -> str:
        """Returns file path stored on the file server."""
        return os.path.join(self.base_dir, path)

    def get_random_filename(self) -> str:
        """Returns a random hash for the file name."""
        return uuid.uuid4().hex

    def upload_file(self, file, id):
        """Upload a file while tracking its progress."""
        # 1: Track metadata about the file
        filename = id
        uploaded_file = id
        meta_file = uploaded_file + ".meta"
        self.uploaded_files[filename] = {"progress": (0, None), "done": False}

        # 2: Create a stream and write bytes of
        # the file to the disk under `uploaded_file` path.
        with open(self.get_filepath(uploaded_file), "wb") as out_file:
            content = file.read(self.chunk_size)
            while content:
                # 2.1 Write the file bytes
                size = out_file.write(content)

                # 2.2 Update the progress metadata
                self.uploaded_files[filename]["progress"] = (
                    self.uploaded_files[filename]["progress"][0] + size,
                    None,
                )
                # 4: Read next chunk of data
                content = file.read(self.chunk_size)

        # 3: Update metadata that the file has been uploaded.
        full_size = self.uploaded_files[filename]["progress"][0]
        self.drive.put(self.get_filepath(uploaded_file))
        self.uploaded_files[filename] = {
            "progress": (full_size, full_size),
            "done": True,
            "uploaded_file": uploaded_file,
        }

        # 4: Write down the metadata about the file to the disk
        meta = {
            "original_path": filename,
            "display_name": os.path.splitext(filename)[0],
            "size": full_size,
            "drive_path": uploaded_file,
        }
        with open(self.get_filepath(meta_file), "wt") as f:
            json.dump(meta, f)

        # 5: Put the file to the drive.
        # It means other components can access get or list them.
        self.drive.put(self.get_filepath(meta_file))
        return meta

    def list_files(self, file_path: str):
        # 1: Get the local file path of the file server.
        file_path = self.get_filepath(file_path)

        # 2: If the file exists in the drive, transfer it locally.
        if not os.path.exists(file_path):
            self.drive.get(file_path)

        if os.path.isdir(file_path):
            result = set()
            for _, _, f in os.walk(file_path):
                for file in f:
                    if not file.endswith(".meta"):
                        for filename, meta in self.uploaded_files.items():
                            if meta["uploaded_file"] == file:
                                result.add(filename)
            return {"asset_names": [v for v in result]}

        # 3: If the filepath is a tar or zip file, list their contents
        if zipfile.is_zipfile(file_path):
            with zipfile.ZipFile(file_path, "r") as zf:
                result = zf.namelist()
        elif tarfile.is_tarfile(file_path):
            with tarfile.TarFile(file_path, "r") as tf:
                result = tf.getnames()
        else:
            raise ValueError("Cannot open archive file!")

        # 4: Returns the matching files.
        return {"asset_names": result}

    def run(self):
        # 1: Imports flask requirements.
        import flask
        from flask import Flask, request
        from flask_cors import CORS

        # 2: Create a flask app
        flask_app = Flask(__name__)
        CORS(flask_app)

        # 3: Define the upload file endpoint
        @flask_app.route("/uploadfile/<id>", methods=["PUT"])
        def upload_file(id: str):
            try:
                """Upload a file directly as form data."""
                f = request.files["file"]
                self.upload_file(f, id)
                resp = flask.Response("File uploaded successfully.")
                return resp
            except Exception:
                trace = str(traceback.print_exc())
                _logger.error(trace)
                resp = flask.Response(trace)
                resp.status_code = 505
                return resp

        @flask_app.get("/")
        def list_files():
            return self.list_files(str(Path(self.base_dir).resolve()))

        # 5: Start the flask app while providing the `host` and `port`.
        flask_app.run(host=self.host, port=self.port, load_dotenv=False)

    def alive(self):
        """Hack: Returns whether the server is alive."""
        return self.url != ""



import argparse
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import requests
from lightning.app.storage.copier import copy_files
from lightning.app.storage.path import shared_storage_path
from lightning.app.utilities.commands import ClientCommand
from lightning.app.utilities.network import LightningClient
from lightning_app.utilities.cloud import _get_project
from pydantic import BaseModel

from lightning_hpo.commands.artefacts.show import _collect_artefact_paths, _filter_paths, ShowArtefactsConfig


class DownloadArtefactsConfig(BaseModel):
    include: Optional[str] = None
    exclude: Optional[str] = None


class DownloadArtefactsCommand(ClientCommand):
    def run(self) -> None:
        # 1. Parse the user arguments.
        parser = argparse.ArgumentParser()
        parser.add_argument("output_dir", type=str, help="Provide the output directory for the artefacts..")
        parser.add_argument("--include", type=str, default=None, help="Provide a regex to include some specific files.")
        parser.add_argument("--exclude", type=str, default=None, help="Provide a regex to exclude some specific files.")
        parser.add_argument(
            "--overwrite", type=str, default=False, help="Whether to overwrite the artefacts if they exist."
        )
        hparams = parser.parse_args()

        output_dir = Path(hparams.output_dir).resolve()

        if not output_dir.exists():
            raise FileNotFoundError(f"The provided directory {output_dir} doesn't exist.")

        config = DownloadArtefactsConfig(include=hparams.include, exclude=hparams.exclude)
        response: List[str] = self.invoke_handler(config=config)

        if not response:
            return

        if len(response[0]) == 2:
            for path, url in response:
                if path.startswith("/"):
                    path = path[1:]
                resp = requests.get(url, allow_redirects=True)
                target_file = Path(os.path.join(output_dir, path))
                with open(target_file, "wb") as f:
                    f.write(resp.content)
        else:
            for path in response:
                source_path = Path(path).resolve()
                shared_storage = shared_storage_path()
                cleaned_path = (
                    str(path)
                    .replace(str(shared_storage) + "/", "")
                    .replace("/artifacts/", "")
                    .replace(os.getcwd() + "/", "")
                )

                target_file = Path(os.path.join(output_dir, cleaned_path))
                if not target_file.parent.exists():
                    os.makedirs(str(target_file.parent), exist_ok=True)
                copy_files(source_path, target_file)

        print("All the specified artifacts were downloaded.")


def _collect_artefact_urls(config: DownloadArtefactsConfig) -> List[Union[str, Tuple[str, str]]]:
    """This function is responsible to collect the files from the shared filesystem."""

    use_localhost = "LIGHTNING_APP_STATE_URL" not in os.environ

    if use_localhost:
        return _collect_artefact_paths(
            config=ShowArtefactsConfig(include=config.include, exclude=config.exclude),
            replace=False,
        )
    else:
        client = LightningClient()
        project_id = _get_project(client).project_id
        app_id = os.getenv("LIGHTNING_CLOUD_APP_ID")
        response = client.lightningapp_instance_service_list_lightningapp_instance_artifacts(project_id, app_id)
        paths = [artifact.filename for artifact in response.artifacts]
        maps = {artifact.filename: artifact.url for artifact in response.artifacts}
        filtered_paths = _filter_paths(paths, include=config.include, exclude=config.exclude)
        return [(path, maps[path]) for path in filtered_paths]



from typing import Optional

from sqlmodel import Field, SQLModel

from lightning_hpo.utilities.enum import Status


class TensorboardConfig(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    sweep_id: str
    shared_folder: str
    status: str = Status.NOT_STARTED
    desired_state: str = Status.RUNNING



import os
from typing import List, Optional, Type

from fastapi import FastAPI
from lightning import BuildConfig, LightningWork
from lightning.app.storage import Path
from sqlmodel import create_engine, select, Session, SQLModel
from uvicorn import run

from lightning_hpo.components.servers.db.models import GeneralModel

engine = None


def general_get(config: GeneralModel):
    with Session(engine) as session:
        statement = select(config.data_cls)
        results = session.exec(statement)
        return results.all()


def general_post(config: GeneralModel):
    with Session(engine) as session:
        data = config.convert_to_model()
        session.add(data)
        session.commit()
        session.refresh(data)
        return data


def general_put(config: GeneralModel):
    with Session(engine) as session:
        assert config.id
        update_data = config.convert_to_model()
        identifier = getattr(update_data.__class__, config.id, None)
        statement = select(update_data.__class__).where(identifier == getattr(update_data, config.id))
        results = session.exec(statement)
        result = results.one()
        for k, v in vars(update_data).items():
            if k in ("id", "_sa_instance_state"):
                continue
            if getattr(result, k) != v:
                setattr(result, k, v)
        session.add(result)
        session.commit()
        session.refresh(result)


def general_delete(config: GeneralModel):
    with Session(engine) as session:
        assert config.id
        update_data = config.convert_to_model()
        identifier = getattr(update_data.__class__, config.id, None)
        statement = select(update_data.__class__).where(identifier == getattr(update_data, config.id))
        results = session.exec(statement)
        result = results.one()
        session.delete(result)
        session.commit()


class Database(LightningWork):
    def __init__(
        self,
        db_file_name: str = "database.db",
        debug: bool = False,
        models: Optional[List[Type[SQLModel]]] = None,  # Just meant to be imported.
    ):
        super().__init__(parallel=True, cloud_build_config=BuildConfig(["sqlmodel"]))
        self.db_file_name = Path(db_file_name)
        self.debug = debug
        self._models = models

    def run(self):
        global engine

        app = FastAPI()
        engine = create_engine(f"sqlite:///{self.db_file_name}", echo=self.debug)

        @app.on_event("startup")
        def on_startup():
            print(f"Creating the following tables {self._models}")
            SQLModel.metadata.create_all(engine)

        app.get("/general/")(general_get)
        app.post("/general/")(general_post)
        app.put("/general/")(general_put)
        app.delete("/general/")(general_delete)

        run(app, host=self.host, port=self.port, log_level="error")

    def alive(self):
        """Hack: Returns whether the server is alive."""
        return self.db_url != ""

    @property
    def db_url(self) -> Optional[str]:
        use_localhost = "LIGHTNING_APP_STATE_URL" not in os.environ
        if use_localhost:
            return self.url
        if self.internal_ip != "":
            return f"http://{self.internal_ip}:{self.port}"
        return self.internal_ip



from typing import List

from lightning import CloudCompute

from lightning_hpo.commands.notebook.run import NotebookConfig, RunNotebookCommand
from lightning_hpo.commands.notebook.show import ShowNotebookCommand
from lightning_hpo.commands.notebook.stop import StopNotebookCommand, StopNotebookConfig
from lightning_hpo.components.notebook import JupyterLab
from lightning_hpo.controllers.controller import Controller
from lightning_hpo.utilities.enum import Status


class NotebookController(Controller):

    model = NotebookConfig

    def on_reconcile_start(self, configs: List[NotebookConfig]):
        for config in configs:
            if config.name not in self.resources and config.desired_state == Status.RUNNING:
                self.resources[config.name] = JupyterLab(
                    kernel="python",
                    cloud_compute=CloudCompute(name=config.cloud_compute),
                    config=config,
                )

    def run_notebook(self, config: NotebookConfig) -> str:
        if config.name in self.resources:
            return f"The notebook `{config.name}` already exists."
        self.db.post(config)
        return f"The notebook `{config.name}` has been created."

    def stop_notebook(self, config: StopNotebookConfig) -> str:
        matched_notebook = None
        for notebook_name, notebook in self.resources.items():
            if notebook_name == config.name:
                matched_notebook = notebook

        if matched_notebook:
            if matched_notebook._config.status != Status.STOPPED:
                notebook: JupyterLab = self.resources[config.name]
                notebook.stop()
                notebook._config.desired_state = notebook._config.status = Status.STOPPED
                self.db.put(notebook._config)
                return f"The notebook `{config.name}` has been stopped."
            return f"The notebook `{config.name}` is already stopped."
        return f"The notebook `{config.name}` doesn't exist."

    def show_notebook(self):
        return self.db.get()

    def configure_commands(self) -> List:
        return [
            {"run notebook": RunNotebookCommand(self.run_notebook)},
            {"stop notebook": StopNotebookCommand(self.stop_notebook)},
            {"show notebooks": ShowNotebookCommand(self.show_notebook)},
        ]



import os
import time
from subprocess import Popen
from uuid import uuid4

from lightning import LightningWork
from lightning.app.storage import Drive
from lightning.app.storage.path import filesystem
from lightning.app.utilities.component import _is_work_context


class Tensorboard(LightningWork):
    def __init__(self, *args, component_name: str, drive: Drive, sleep: int = 5, **kwargs):
        super().__init__(*args, **kwargs)
        self.component_name = component_name
        self.drive = drive
        self.sleep = sleep

    def run(self):
        local_folder = f"./tensorboard_logs/{uuid4()}"
        self.drive.component_name = self.component_name

        os.makedirs(local_folder, exist_ok=True)

        # Note: Used tensorboard built-in sync methods but it doesn't seem to work.
        cmd = f"tensorboard --logdir={local_folder} --host {self.host} --port {self.port}"
        self._process = Popen(cmd, shell=True, env=os.environ)

        fs = filesystem()

        while True:
            fs.invalidate_cache()
            fs.get(str(self.drive.root), local_folder)
            time.sleep(self.sleep)

    def on_exit(self):
        if _is_work_context():
            assert self._process
            self._process.kill()

    @property
    def updates(self):
        return []



from lightning_hpo.components.servers.db.connector import DatabaseConnector  # noqa F401
from lightning_hpo.components.servers.db.server import Database  # noqa F401
from lightning_hpo.components.servers.db.visualization import DatabaseViz  # noqa F401



import os.path as ops

from lightning import LightningApp

from lightning_hpo import HPOCloudCompute, Sweep
from lightning_hpo.algorithm.optuna import OptunaAlgorithm
from lightning_hpo.distributions.distributions import Categorical, IntUniform, LogUniform, Uniform

app = LightningApp(
    Sweep(
        script_path=ops.join(ops.dirname(__file__), "scripts/train.py"),
        n_trials=5,
        simultaneous_trials=2,
        distributions={
            "model.lr": LogUniform(0.001, 0.1),
            "model.gamma": Uniform(0.5, 0.8),
            "data.batch_size": Categorical([16, 32, 64]),
            "trainer.max_epochs": IntUniform(1, 5),
        },
        algorithm=OptunaAlgorithm(direction="maximize"),
        cloud_compute=HPOCloudCompute("gpu-fast-multi", count=2),  # 2 * 4 V100
        framework="pytorch_lightning",
        logger="wandb",
        sweep_id="Optimizing a Simple CNN over MNIST with Lightning HPO",
    )
)



import logging
from typing import Any, Dict, List, Optional, Tuple

import optuna
from optuna import create_study, Study, Trial
from optuna.distributions import (
    BaseDistribution,
    CategoricalDistribution,
    IntUniformDistribution,
    LogUniformDistribution,
    UniformDistribution,
)

from lightning_hpo.algorithm.base import Algorithm
from lightning_hpo.commands.sweep.run import TrialConfig
from lightning_hpo.distributions import DistributionDict

_logger = logging.getLogger(__name__)

_DISTRIBUTION_TO_OPTUNA = {
    "uniform": UniformDistribution,
    "int_uniform": IntUniformDistribution,
    "log_uniform": LogUniformDistribution,
    "categorical": CategoricalDistribution,
}


class OptunaAlgorithm(Algorithm):
    def __init__(self, study: Optional[Study] = None, direction: Optional[str] = "minimize") -> None:
        self.study = study or create_study(direction=direction)
        self.trials: Dict[int, Trial] = {}
        self.reports = {}
        self.distributions: Dict[str, BaseDistribution] = {}

    def register_distributions(self, distributions: Dict[str, DistributionDict]):
        for var_name, distribution in distributions.items():
            distribution_cls = _DISTRIBUTION_TO_OPTUNA[distribution["distribution"]]
            distribution = distribution_cls(**distribution["params"])
            self.distributions[var_name] = distribution

    def register_trials(self, trials_config: List[TrialConfig]) -> None:
        for trial_config in trials_config:
            trial = optuna.trial.create_trial(
                params=trial_config.params.params, distributions=self.distributions, value=trial_config.best_model_score
            )
            self.study.add_trial(trial)

    def trial_start(self, trial_id: int) -> None:
        if trial_id not in self.trials:
            self.trials[trial_id] = self.study.ask(self.distributions)

    def trial_end(self, trial_id: int, score: float):
        self.study.tell(trial_id, score)

        _logger.info(
            f"Trial {trial_id} finished with value: {score} and parameters: {self.trials[trial_id].params}. "  # noqa: E501
            f"Best is trial {self.study.best_trial.number} with value: {self.study.best_trial.value}."
        )

    def should_prune(self, trial_id: int, reports: List[Tuple[float, int]]) -> bool:
        trial = self.trials[trial_id]
        if trial_id not in self.reports:
            self.reports[trial_id] = []

        for report in reports:
            if report in self.reports[trial_id]:
                continue
            trial.report(*report)
            self.reports[trial_id].append(report)

            if trial.should_prune():
                _logger.info(f"Trial {trial_id} pruned.")
                return True

        return False

    def get_params(self, trial_id: int) -> Dict[str, Any]:
        params = self.trials[trial_id].params
        out = {}
        for k, v in params.items():
            if v == int(v):
                out[k] = int(v)
            else:
                out[k] = v
        return out



import sys
from subprocess import Popen

from lightning import BuildConfig, LightningWork


class DatabaseViz(LightningWork):
    def __init__(
        self,
        db_file_name: str = "database.db",
    ):
        super().__init__(parallel=True, cloud_build_config=BuildConfig(["sqlite-web"]))
        self.db_file_name = db_file_name

    def run(self):
        cmd = f"sqlite_web {self.db_file_name} -H {self.host} -p {self.port}"
        Popen(cmd, shell=True, stdout=sys.stdout, stderr=sys.stderr).wait()

    def alive(self):
        """Hack: Returns whether the server is alive."""
        return self.url != ""



from typing import Any, Dict, Optional

from lightning.app.components.python import TracerPythonScript
from lightning.app.components.training import LightningTrainingComponent, PyTorchLightningScriptRunner

from lightning_hpo.framework.agnostic import Objective


class PyTorchLightningObjective(Objective, PyTorchLightningScriptRunner):

    """This component executes a PyTorch Lightning script
    and injects a callback in the Trainer at runtime in order to start tensorboard server."""

    def __init__(self, *args, logger: str, sweep_id: str, trial_id: int, num_nodes: int, **kwargs):
        Objective.__init__(self, logger=logger, sweep_id=sweep_id, trial_id=trial_id, **kwargs)
        PyTorchLightningScriptRunner.__init__(self, *args, num_nodes=num_nodes, **kwargs)

    def configure_tracer(self):
        if self.node_rank == 0:
            return Objective.configure_tracer(self)
        return TracerPythonScript.configure_tracer(self)

    def run(self, params: Optional[Dict[str, Any]] = None, restart_count: int = 0, **kwargs):
        self.params = params
        return PyTorchLightningScriptRunner.run(self, params=params, **kwargs)

    def on_after_run(self, script_globals):
        PyTorchLightningScriptRunner.on_after_run(self, script_globals)

    @classmethod
    def distributions(cls):
        return None


class ObjectiveLightningTrainingComponent(LightningTrainingComponent):
    def __init__(self, *args, trial_id: int, logger: str, sweep_id: str, num_nodes: int = 1, **kwargs):
        super().__init__(
            *args,
            script_runner=PyTorchLightningObjective,
            logger=logger,
            sweep_id=sweep_id,
            trial_id=trial_id,
            num_nodes=num_nodes,
            **kwargs,
        )
        self.trial_id = trial_id
        self.has_stopped = False
        self.pruned = False
        self.params = None
        self.restart_count = 0
        self.sweep_id = sweep_id
        self.reports = []
        self.has_stored = False

    def run(self, params: Optional[Dict[str, Any]] = None, restart_count: int = 0):
        self.params = params
        self.restart_count = restart_count
        super().run(params=params, restart_count=restart_count)

    @property
    def monitor(self):
        return self.ws[0].monitor

    @property
    def best_model_path(self):
        return self.ws[0].best_model_path

    @property
    def best_model_score(self):
        return self.ws[0].best_model_score

    @property
    def has_failed(self) -> bool:
        return any(w.has_failed for w in self.works())

    def stop(self):
        for w in self.works():
            w.stop()
        self.has_stopped = True

    @classmethod
    def distributions(cls):
        return {}



from time import sleep
from typing import List

import requests
from lightning.app.frontend import StreamlitFrontend
from lightning.app.storage import Drive
from lightning.app.structures import Dict

from lightning_hpo import Sweep
from lightning_hpo.commands.sweep.delete import DeleteSweepCommand, DeleteSweepConfig
from lightning_hpo.commands.sweep.run import RunSweepCommand, SweepConfig
from lightning_hpo.commands.sweep.show import ShowSweepsCommand
from lightning_hpo.commands.sweep.stop import StopSweepCommand, StopSweepConfig
from lightning_hpo.commands.tensorboard.stop import TensorboardConfig
from lightning_hpo.components.servers.db.models import GeneralModel
from lightning_hpo.controllers.controller import Controller
from lightning_hpo.loggers import LoggerType
from lightning_hpo.utilities.enum import Status


class SweepController(Controller):

    model = SweepConfig
    model_id = "sweep_id"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tensorboard_sweep_id = None

    def on_reconcile_start(self, sweeps: List[SweepConfig]):
        # 1: Retrieve the tensorboard configs from the database
        if self.tensorboard_sweep_id is None:
            self.tensorboard_sweep_id = [c.sweep_id for c in self.db.get(TensorboardConfig)]

        # 2: Create the Sweeps
        for sweep in sweeps:
            id = sweep.sweep_id
            if sweep.logger == LoggerType.TENSORBOARD.value and id not in self.tensorboard_sweep_id:
                self.tensorboard_sweep_id.append(id)
                drive = Drive(f"lit://{id}", component_name="logs")
                self.db.post(TensorboardConfig(sweep_id=id, shared_folder=str(drive.root)), "id")

            if id not in self.resources:
                self.resources[id] = Sweep.from_config(
                    sweep,
                    code={"drive": self.drive, "name": id},
                )

    def on_reconcile_end(self, updates: List[SweepConfig]):
        for update in updates:
            if update.status == Status.SUCCEEDED:
                for w in self.resources[update.sweep_id].works():
                    w.stop()
                self.resources.pop(update.sweep_id)

    def run_sweep(self, config: SweepConfig) -> str:
        sweep_ids = list(self.resources.keys())
        if config.sweep_id not in sweep_ids:
            self.db.post(config)
            return f"Launched a sweep {config.sweep_id}"
        return f"The current Sweep {config.sweep_id} is running. It couldn't be updated."

    def show_sweeps(self) -> List[Dict]:
        return [sweep.json() for sweep in self.db.get()]

    def stop_sweep(self, config: StopSweepConfig):
        sweep_ids = list(self.resources.keys())
        if config.sweep_id in sweep_ids:
            # TODO: Add support for __del__ in lightning
            sweep: Sweep = self.resources[config.sweep_id]
            for w in sweep.works():
                w.stop()
            sweep_config = sweep._sweep_config
            sweep_config.status = Status.STOPPED
            for trial in sweep_config.trials.values():
                if trial.status != Status.SUCCEEDED:
                    trial.status = Status.STOPPED
            self.db.put(sweep_config)
            return f"Stopped the sweep `{config.sweep_id}`"
        return f"We didn't find the sweep `{config.sweep_id}`"

    def delete_sweep(self, config: DeleteSweepConfig):
        sweep_ids = list(self.resources.keys())
        if config.sweep_id in sweep_ids:
            sweep: Sweep = self.resources[config.sweep_id]
            for w in sweep.works():
                w.stop()
            sweep_config = sweep._sweep_config
            self.db.delete(sweep_config)
            del self.resources[config.sweep_id]
            return f"Deleted the sweep `{config.sweep_id}`"
        return f"We didn't find the sweep `{config.sweep_id}`"

    def configure_commands(self):
        return [
            {"delete sweep": DeleteSweepCommand(self.delete_sweep)},
            {"run sweep": RunSweepCommand(self.run_sweep)},
            {"show sweeps": ShowSweepsCommand(self.show_sweeps)},
            {"stop sweep": StopSweepCommand(self.stop_sweep)},
        ]

    def configure_layout(self):
        return StreamlitFrontend(render_fn=render_fn)


def render_fn(state):
    import streamlit as st
    import streamlit.components.v1 as components

    if not state.db_url:
        sleep(1)
        st.experimental_rerun()

    resp = requests.get(state.db_url + "/general/", data=GeneralModel.from_cls(SweepConfig).json())
    sweeps: List[SweepConfig] = resp.json()

    if not sweeps:
        st.header("You haven't launched any sweeps yet.")
        st.write("Here is an example to submit a sweep.")
        st.code(
            'lightning run sweep train.py --n_trials=2 --num_nodes=2 --model.lr="log_uniform(0.001, 0.1)" --trainer.max_epochs=5 --trainer.callbacks=ModelCheckpoint'
        )
        return

    user_sweeps = {}
    for sweep in sweeps:
        username, sweep_id = sweep["sweep_id"].split("-")
        if username not in user_sweeps:
            user_sweeps[username] = {}
        user_sweeps[username][sweep_id] = sweep

    user_tabs = st.tabs(user_sweeps)
    for tab, username in zip(user_tabs, user_sweeps):
        with tab:
            for sweep_id, sweep in user_sweeps[username].items():
                with st.expander(f"{sweep_id} / {sweep['status']}"):
                    trials_tab, logging_tab = st.tabs(["Trials", "Logging"])
                    with trials_tab:
                        for trial_id, trial in sweep["trials"].items():
                            if st.checkbox(f"Trial {trial_id}", key=f"checkbox_{trial_id}_{sweep_id}"):
                                st.json(
                                    {
                                        "params": trial["params"]["params"],
                                        "monitor": trial["monitor"],
                                        "best_model_score": trial["best_model_score"],
                                    }
                                )
                            with logging_tab:
                                url = trial.get("url", None)
                                if url:
                                    components.html(
                                        f'<a href="{url}" target="_blank">Weights & Biases URL</a>', height=50
                                    )



from subprocess import Popen
from typing import Optional

from lightning.app.utilities.component import _is_work_context
from lit_jupyter import JupyterLab

from lightning_hpo.commands.notebook.run import NotebookConfig
from lightning_hpo.utilities.enum import Status


class JupyterLab(JupyterLab):
    def __init__(self, *args, config: NotebookConfig, **kwargs):
        super().__init__(*args, **kwargs)
        self._config = config
        self.ready = False
        self.has_updated = False
        self._process: Optional[Popen] = None

    def run(self, *args, **kwargs):
        super().run()
        self.ready = True
        self.has_updated = True

    # TODO: Cleanup exit mechanism in lightning.
    def on_exit(self):
        if _is_work_context():
            assert self._process
            self._process.kill()

    @property
    def updates(self):
        if self.has_updated and self.ready:
            self._config.status = Status.RUNNING
            self.has_updated = False
            return [self._config]
        return []
